/*
 * @(#)HextrisN.c
 *
 * Copyright 1993 - 2008  David A. Bagley, bagleyd@tux.org
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Next methods file for Hextris */
#include "file.h"
#include "rngs.h"
#include "HextrisNP.h"

static Point hexagon_unit[MAX_ORIENT][6] = {
	{ {0, 0}, {2, 0}, {1, 1}, {-1, 1}, {-2, 0}, {-1, -1} },
	{ {0, 0}, {1, 1}, {0, 2}, {-1, 1}, {-1, -1}, {0, -2} }
};
static Point hexagon_list[MAX_ORIENT][7];
static Point hexagon_border[7];

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "whextris.ini"
#endif

#define SECTION "setup"

#else

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static void resizeNextGame(NextHextrisWidget w);
static void destroyNextGame(Widget old);
#if 0
static void sizeNextGame(NextHextrisWidget w);
#endif
static Boolean setValuesNextGame(Widget current, Widget request, Widget renew);
static void quitNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void initializeNextGame(Widget request, Widget renew);
static void exposeNextGame(Widget renew, XEvent *event, Region region);
static void hideNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void enterNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void leaveNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);
static void showNextNextGame(NextHextrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsNextGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <KeyPress>n: ShowNext()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

static XtActionsRec actionsListNextGame[] =
{
	{(char *) "Quit", (XtActionProc) quitNextGame},
	{(char *) "Hide", (XtActionProc) hideNextGame},
	{(char *) "ShowNext", (XtActionProc) showNextNextGame},
	{(char *) "Enter", (XtActionProc) enterNextGame},
	{(char *) "Leave", (XtActionProc) leaveNextGame}
};

static XtResource resourcesNextGame[] =
{
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(NextHextrisWidget, core.width),
	 XtRString, (caddr_t) "146"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(NextHextrisWidget, core.height),
	 XtRString, (caddr_t) "111"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextHextrisWidget, hextris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(NextHextrisWidget, hextris.background),
	 XtRString, (caddr_t) XtDefaultBackground},
	{XtNcolumns, XtCColumns, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, hextris.columns),
	 XtRString, (caddr_t) "10"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, hextris.units),
	 XtRString, (caddr_t) "4"},
	{XtNcorners, XtCCorners, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.corners),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, nextHextris.newBonus),
	 XtRString, (caddr_t) "FALSE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, hextris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(NextHextrisWidget, hextris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNrandomNumber, XtCRandomNumber, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, nextHextris.randomNumber),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomRotation, XtCRandomRotation, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, nextHextris.randomRotation),
	 XtRString, (caddr_t) "-1"},
	{XtNrandomReflection, XtCRandomReflection, XtRInt, sizeof (int),
	 XtOffset(NextHextrisWidget, nextHextris.randomReflection),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(NextHextrisWidget, hextris.select),
	 XtRCallback, (caddr_t) NULL}
};

NextHextrisClassRec nextHextrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "NextHextris",	/* class name */
		sizeof (NextHextrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) initializeNextGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListNextGame,	/* actions */
		XtNumber(actionsListNextGame),	/* num actions */
		resourcesNextGame,	/* resources */
		XtNumber(resourcesNextGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) destroyNextGame,	/* destroy */
		(XtWidgetProc) resizeNextGame,	/* resize */
		(XtExposeProc) exposeNextGame,	/* expose */
		(XtSetValuesFunc) setValuesNextGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsNextGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass nextHextrisWidgetClass = (WidgetClass) & nextHextrisClassRec;

static void
setNextGame(NextHextrisWidget w, int reason)
{
	hextrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}
#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static void
checkThings(NextHextrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->hextris.units < MIN_UNITS) {
		w->hextris.units = MIN_UNITS;
		w->hextris.next.units = w->hextris.units;
		intCat(&buf1, "Number of units out of bounds, use at least ",
			MIN_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->hextris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->hextris.units > MAX_UNITS) {
		w->hextris.units = MAX_UNITS;
		w->hextris.next.units = w->hextris.units;
		intCat(&buf1, "Number of units out of bounds, use at most ",
			MAX_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->hextris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
}

static void
draw3DFrame(const NextHextrisWidget w, Boolean inside,
	int x, int y, int width, int height)
{
	GC gc;

	if (inside)
		gc = w->hextris.inverseGC[3];
	else
		gc = w->hextris.inverseGC[1];
	FILLRECTANGLE(w, dr, gc,
		x, y, width, 1);
	FILLRECTANGLE(w, dr, gc,
		x, y + 1, 1, height - 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 1, width - 2, 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 2, 1, height - 3);
	if (inside)
		gc = w->hextris.inverseGC[0];
	else
		gc = w->hextris.inverseGC[1];
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + height - 1, width - 1, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 1, y + 1, 1, height - 2);
	FILLRECTANGLE(w, dr, gc,
		x + 2, y + height - 2, width - 3, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 2, y + 2, 1, height - 4);
}

static void
eraseFrame(NextHextrisWidget w)
{
	FILLRECTANGLE(w, dr, (w->hextris.showNext) ?
		w->hextris.inverseGC[2] : w->hextris.inverseGC[1],
		w->hextris.offset.x, w->hextris.offset.y,
		w->hextris.viewWidth, w->hextris.viewHeight);
}

static void
drawFrame(NextHextrisWidget w)
{
	int offsetX = w->hextris.offset.x - 2;
	int offsetY = w->hextris.offset.y - 2;
	int lengthX = w->hextris.offset.x + w->hextris.viewWidth + 2;
	int lengthY = w->hextris.offset.y + w->hextris.viewHeight + 2;

	if (w->hextris.showNext) {
		FILLRECTANGLE(w, dr, w->hextris.inverseGC[4],
			0, 0, w->core.width, offsetY);
		FILLRECTANGLE(w, dr, w->hextris.inverseGC[4],
			0, offsetY, offsetX, lengthY - offsetY);
		FILLRECTANGLE(w, dr, w->hextris.inverseGC[4],
			lengthX, offsetY,
			w->core.width - lengthX, lengthY - offsetY);
		FILLRECTANGLE(w, dr, w->hextris.inverseGC[4],
			0, lengthY, w->core.width, offsetY + 1);
		draw3DFrame(w, w->hextris.showNext,
			offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	} else {
		FILLRECTANGLE(w, dr, w->hextris.inverseGC[1],
			w->hextris.offset.x, w->hextris.offset.y,
			w->hextris.viewWidth, w->hextris.viewHeight);
	}
}

static void
drawUnitBorder(NextHextrisWidget w, int i, int j)
{
	if (i >= 0 && i < MAX_UNITS &&
			j >= 0 && j < MAX_UNITS + 1) {
		GC gc = w->hextris.inverseGC[2];

		hexagon_border[0].x = i * w->hextris.hexWidth +
			w->hextris.offset.x - 1 +
			(((j + 1 + w->hextris.nextThing.size) % 2) ?
			0 : (w->hextris.hexWidth + 1) / 2);
		hexagon_border[0].y = (j - 1) * w->hextris.hexHeight +
			w->hextris.offset.y;
		POLYLINE(w, dr, gc,
			hexagon_border, MAX_SIDES + 1, False);
	}
}

static void
drawUnit(NextHextrisWidget w, int pmid, int cid, int i, int j)
{
	if (i >= 0 && i < MAX_UNITS &&
			j >= 0 && j < MAX_UNITS + 1) {
		hexagon_border[0].x = i * w->hextris.hexWidth +
			w->hextris.offset.x - 1 +
			(((j + 1 + w->hextris.nextThing.size) % 2) ?
			0 : (w->hextris.hexWidth + 1) / 2);
		hexagon_border[0].y = (j - 1) * w->hextris.hexHeight +
			w->hextris.offset.y;
		if (w->hextris.next.corners) {
			fillPolygon((HextrisWidget) w, pmid, cid,
				hexagon_border, True, False);
		} else {
			Point hexagon_borderTrunc[MAX_SIDES + 1];
			Point aPoint, bPoint, aPointTrunc, bPointTrunc;
			int k;

			aPoint.x = aPoint.y = 0;
			bPoint.x = bPoint.y = 0;
			aPointTrunc.x = aPointTrunc.y = 0;
			bPointTrunc.x = -hexagon_border[1].x / 2;
			bPointTrunc.y = hexagon_border[1].y / 2;
			for (k = 0; k < MAX_SIDES; k++) {
				aPoint.x += hexagon_border[k + 1].x;
				aPoint.y += hexagon_border[k + 1].y;
				aPointTrunc.x = (aPoint.x + bPoint.x) / 2;
				aPointTrunc.y = (aPoint.y + bPoint.y) / 2;
				hexagon_borderTrunc[k + 1].x =
					aPointTrunc.x - bPointTrunc.x;
				hexagon_borderTrunc[k + 1].y =
					aPointTrunc.y - bPointTrunc.y;
				bPoint = aPoint;
				bPointTrunc = aPointTrunc;
			}
			hexagon_borderTrunc[0].x = hexagon_border[0].x -
				hexagon_border[1].x / 2;
			hexagon_borderTrunc[0].y = hexagon_border[0].y +
				hexagon_border[1].y / 2 + 1;
			fillPolygon((HextrisWidget) w, pmid, cid,
				hexagon_borderTrunc, False, False);
		}
	}
}

#if 0
static void
clearThing(HextrisWidget w)
{
	XFillRectangle(display, mainWin, w->hextris.inverseGC[2],
		topMidX,
		topMidY - (w->hextris.hexHeight * HEX_ROW_COL(HEX_SIZE) +
		w->hextris.hexHeight / 3) / 2 + 3,
		topRWidth, w->hextris.hexHeight * HEX_ROW_COL(HEX_SIZE) +
		w->hextris.hexHeight / 3 + 4);
}
#endif

static void
clearThing(NextHextrisWidget w)
{
	FILLRECTANGLE(w, dr, w->hextris.inverseGC[2],
		w->hextris.offset.x, w->hextris.offset.y,
		w->hextris.viewWidth, w->hextris.viewHeight);
}

/*static void
drawThing(NextHextrisWidget w)
{
	int i, j;

	for (i = 0; i < w->hextris.nextThing.size; i++)
		for (j = 0; j < w->hextris.nextThing.size; j++)
			if (tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyNumber].shape[j][i]) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyNumber].shape),
					w->hextris.nextThing.size),
					w->hextris.nextThing.colorNumber,
					w->hextris.nextThing.xpos + i,
					w->hextris.nextThing.ypos + j);
			}
}*/

static void
drawThing(NextHextrisWidget w)
{
	int i, j;
	int len = HEX_ROW_COL(w->hextris.nextThing.size);

	if (w->hextris.nextThing.polyNumber == IGNORE_RAND)
		return;
	clearThing(w);
	drawFrame(w);
	for (i = 0; i < len; i++)
		for (j = 0; j < len; j++)
			if (inHexagon(w->hextris.nextThing.size, i, j) != 0 &&
					tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyNumber].shape[j][i] != 0) {
				drawUnit(w, polyPixmap(i, j,
					(tris[w->hextris.nextThing.units].polyhex
					[w->hextris.nextThing.polyNumber].shape),
					w->hextris.nextThing.size),
					w->hextris.nextThing.colorNumber,
					w->hextris.nextThing.xpos + i,
					w->hextris.nextThing.ypos + j);
				if (w->hextris.next.corners)
					drawUnitBorder(w,
						w->hextris.nextThing.xpos + i,
						w->hextris.nextThing.ypos + j);
			}
}

static void
redoNextNext(NextHextrisWidget w)
{
	redoNext((HextrisWidget) w);
	w->hextris.nextThing.xpos =
		(MAX_UNITS - w->hextris.nextThing.size) / 2;
	w->hextris.nextThing.ypos =
		(MAX_UNITS - w->hextris.nextThing.size) / 2;
#ifdef DEBUG
	(void) printf("Size: %d, Poly: %d, Color %d\n",
		w->hextris.nextThing.size, w->hextris.nextThing.polyNumber,
		w->hextris.nextThing.colorNumber);
#endif
}

#if 0
static void
resetThings(NextHextrisWidget w)
{
}

static void
resizeThings(NextHextrisWidget w)
{
}
#endif

#ifdef WINVER
static void
setValuesNextGame(NextHextrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->hextris.units = GetPrivateProfileInt(SECTION, "units",
		DEFAULT_UNITS, INIFILE);
	w->hextris.corners = (BOOL) GetPrivateProfileInt(SECTION, "corners",
		DEFAULT_CORNERS, INIFILE);
	w->hextris.mixed = (BOOL) GetPrivateProfileInt(SECTION, "mixed",
		DEFAULT_MIXED, INIFILE);
	w->hextris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->hextris.showNext = (BOOL) GetPrivateProfileInt(SECTION, "showNext",
		FALSE, INIFILE);
	w->hextris.mono = (BOOL) GetPrivateProfileInt(SECTION, "mono",
		DEFAULT_MONO, INIFILE);
	w->hextris.reverse = (BOOL) GetPrivateProfileInt(SECTION,
		"reverseVideo", DEFAULT_REVERSE, INIFILE);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION, "background", "255 255 255",
		szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->hextris.inverseGC[1] = RGB(color.red, color.green, color.blue);
	w->hextris.inverseGC[0] = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->hextris.inverseGC[2] = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->hextris.inverseGC[3] = RGB(darker(darker(color.red)),
		darker(darker(color.green)), darker(darker(color.blue)));
	w->hextris.inverseGC[4] = RGB(darker(darker(darker(color.red))),
		darker(darker(darker(color.green))),
		darker(darker(darker(color.blue))));
	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION, buf,
			polyhexColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->hextris.thingGC[thing][2] =
			RGB(color.red, color.green, color.blue);
		w->hextris.thingGC[thing][1] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->hextris.thingGC[thing][3] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
		w->hextris.thingGC[thing][0] =
			RGB(brighter(brighter(color.red)),
			brighter(brighter(color.green)),
			brighter(brighter(color.blue)));
		w->hextris.thingGC[thing][4] =
			RGB(darker(darker(color.red)),
			darker(darker(color.green)),
			darker(darker(color.blue)));
	}
}

void
destroyNextGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	/*PostQuitMessage(0);*/
}

#else
static void
getColor(NextHextrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->hextris.reverse) {
		values.background = w->hextris.foreground;
	} else {
		values.background = w->hextris.background;
	}
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing][2])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing][2]);
	w->hextris.thingGC[thing][2] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing][2])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing][2]);
	w->hextris.thingGC[thing][2] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				brighter((HextrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing][1])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing][1]);
	w->hextris.thingGC[thing][1] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				darker((HextrisWidget) w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing][3])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing][3]);
	w->hextris.thingGC[thing][3] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				brighter((HextrisWidget) w,
				brighter((HextrisWidget) w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing][0])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing][0]);
	w->hextris.thingGC[thing][0] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->hextris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyhexColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->hextris.thingColor[thing] =
				darker((HextrisWidget) w,
				darker((HextrisWidget) w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyhexColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyhex ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->hextris.reverse) {
			values.background = w->hextris.foreground;
			values.foreground = w->hextris.background;
		} else {
			values.background = w->hextris.background;
			values.foreground = w->hextris.foreground;
		}
	}
	if (w->hextris.thingGC[thing][4])
		XtReleaseGC((Widget) w, w->hextris.thingGC[thing][4]);
	w->hextris.thingGC[thing][4] = XtGetGC((Widget) w, valueMask,
		&values);
}

static void
setAllColors(NextHextrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyhex;

	valueMask = GCForeground | GCBackground;
	if (w->hextris.reverse) {
		values.background = w->hextris.background;
		values.foreground = w->hextris.foreground;
	} else {
		values.foreground = w->hextris.background;
		values.background = w->hextris.foreground;
	}
	if (w->hextris.inverseGC[1])
		XtReleaseGC((Widget) w, w->hextris.inverseGC[1]);
	w->hextris.inverseGC[1] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->hextris.mono) {
		values.foreground = brighter((HextrisWidget) w, (w->hextris.reverse) ?
			w->hextris.foreground : w->hextris.background);
	}
	if (w->hextris.inverseGC[0])
		XtReleaseGC((Widget) w, w->hextris.inverseGC[0]);
	w->hextris.inverseGC[0] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->hextris.mono) {
		values.foreground = darker((HextrisWidget) w, (w->hextris.reverse) ?
			w->hextris.foreground : w->hextris.background);
	}
	if (w->hextris.inverseGC[2])
		XtReleaseGC((Widget) w, w->hextris.inverseGC[2]);
	w->hextris.inverseGC[2] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->hextris.mono) {
		values.foreground = darker((HextrisWidget) w, darker((HextrisWidget) w, (w->hextris.reverse) ?
			w->hextris.foreground : w->hextris.background));
	}
	if (w->hextris.inverseGC[3])
		XtReleaseGC((Widget) w, w->hextris.inverseGC[3]);
	w->hextris.inverseGC[3] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->hextris.mono) {
		values.foreground = darker((HextrisWidget) w, darker((HextrisWidget) w, darker((HextrisWidget) w,
			(w->hextris.reverse) ? w->hextris.foreground :
			w->hextris.background)));
	}
	if (w->hextris.inverseGC[4])
		XtReleaseGC((Widget) w, w->hextris.inverseGC[4]);
	w->hextris.inverseGC[4] = XtGetGC((Widget) w, valueMask, &values);
	for (polyhex = 0; polyhex < MAX_START_POLYHEXES; polyhex++)
		getColor(w, polyhex);
}

static Boolean
setValuesNextGame(Widget current, Widget request, Widget renew)
{
	NextHextrisWidget c = (NextHextrisWidget) current;
	NextHextrisWidget w = (NextHextrisWidget) renew;
	Boolean redraw = False, setColors = False;

	checkThings(w);
	if (w->hextris.showNext != c->hextris.showNext) {
		redraw = True;
	}
	if (w->hextris.reverse != c->hextris.reverse ||
			w->hextris.mono != c->hextris.mono) {
		setAllColors(w);
		redraw = True;
	} else if (w->hextris.background != c->hextris.background ||
			w->hextris.foreground != c->hextris.foreground ||
			setColors) {
		setAllColors(w);
		redraw = True;
	} else if (w->hextris.plainPiece != c->hextris.plainPiece) {
		redraw = True;
	}
#if 0
	if (w->hextris.units != c->hextris.units) {
		w->hextris.next.units = w->hextris.units;
		resetThings(w);
	} else if (w->hextris.corners != c->hextris.corners) {
		w->hextris.next.corners = w->hextris.corners;
		resetThings(w);
	} else if (w->hextris.mixed != c->hextris.mixed) {
		w->hextris.next.mixed = w->hextris.mixed;
		resetThings(w);
	}
#endif
	if (w->hextris.menu != -1) {
		switch (w->hextris.menu) {
		case 3:
			w->hextris.menu = -1;
			clearThing(w);
			redraw = True;
			break;
		default:
			w->hextris.menu = -1;
			break;
		}
	}
	if (w->nextHextris.randomNumber == CLEARNEXT_RAND) {
		if (w->hextris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
		}
		w->nextHextris.randomNumber = IGNORE_RAND;
		redraw = True;
	} else if (w->nextHextris.randomNumber != IGNORE_RAND) {
		w->hextris.nextThing.randomNumber = w->nextHextris.randomNumber;
		w->hextris.nextThing.randomRotation = w->nextHextris.randomRotation;
		w->hextris.nextThing.randomReflection = w->nextHextris.randomReflection;
		w->hextris.nextThing.bonus = w->nextHextris.newBonus;
		w->hextris.next.units = w->hextris.units;
		w->hextris.next.corners = w->hextris.corners;
		w->hextris.next.mixed = w->hextris.mixed;
		redoNextNext(w);
		if (w->hextris.showNext) {
			drawThing(w);
		}
		w->nextHextris.randomNumber = IGNORE_RAND;
		redraw = True;
	}
	return (redraw);
}

static void
quitNextGame(NextHextrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	XtCloseDisplay(display);
	exit(0);
}

static void
destroyNextGame(Widget old)
{
	NextHextrisWidget w = (NextHextrisWidget) old;
	int thing, i;

	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			XtReleaseGC(old, w->hextris.thingGC[thing][i]);
	}
	for (i = 0; i < BG_SHADES; i++)
		XtReleaseGC(old, w->hextris.inverseGC[i]);
	XtRemoveCallbacks(old, XtNselectCallback, w->hextris.select);
}
#endif

#define SQRT2_2 0.70710678 /* sqrt(2) / 2 */

#ifndef WINVER
static
#endif
void
resizeNextGame(NextHextrisWidget w)
{
	int i;
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->hextris.hexHeight = MIN(3 * (w->core.height - 4) /
		(3 * MAX_UNITS + 1),
		(w->core.width - 4) * SQRT2_2 / MAX_UNITS);
	w->hextris.hexWidth = (int) (w->hextris.hexHeight / SQRT2_2);
	w->hextris.viewWidth = w->hextris.hexWidth *
		MAX_UNITS + 2;
	w->hextris.viewHeight = w->hextris.hexHeight *
		 (3 * MAX_UNITS + 1) / 3 + 1;
	w->hextris.offset.x = ((int) w->core.width -
		w->hextris.viewWidth) / 2;
	w->hextris.offset.y = ((int) w->core.height -
		w->hextris.viewHeight) / 2;

	for (i = 0; i < MAX_SIDES; i++) {
		hexagon_list[CORNERS][i].x =
			(w->hextris.hexWidth - 2) / 2 *
			hexagon_unit[CORNERS][i].x;
		hexagon_list[CORNERS][i].y =
			(2 * w->hextris.hexHeight / 3 - 1) / 2 *
			hexagon_unit[CORNERS][i].y;
		hexagon_list[NOCORN][i].x =
			(w->hextris.hexWidth - 1) / 4 *
			hexagon_unit[NOCORN][i].x;
		hexagon_list[NOCORN][i].y =
			3 * (2 * w->hextris.hexHeight / 3 - 1) / 4 *
			hexagon_unit[NOCORN][i].y;
		hexagon_border[i].x =
			w->hextris.hexWidth / 2 * hexagon_unit[CORNERS][i].x;
		hexagon_border[i].y =
			w->hextris.hexHeight / 3 * hexagon_unit[CORNERS][i].y;
	}
	hexagon_list[CORNERS][MAX_SIDES].x =
		-hexagon_list[CORNERS][MAX_SIDES / 2].x;
	hexagon_list[CORNERS][MAX_SIDES].y =
		-hexagon_list[CORNERS][MAX_SIDES / 2].y;
	hexagon_list[NOCORN][MAX_SIDES].x =
		-hexagon_list[NOCORN][MAX_SIDES / 2].x;
	hexagon_list[NOCORN][MAX_SIDES].y =
		-hexagon_list[NOCORN][MAX_SIDES / 2].y;
	hexagon_border[MAX_SIDES].x = -hexagon_border[MAX_SIDES / 2].x;
	hexagon_border[MAX_SIDES].y = -hexagon_border[MAX_SIDES / 2].y;
}

#if 0
#ifndef WINVER
static
#endif
void
6izeNextGame(NextHextrisWidget w)
{
	resetThings(w);
	resizeNextGame(w);
}
#endif

#ifndef WINVER
static
#endif
void
initializeNextGame(
#ifdef WINVER
NextHextrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int corners, mixed, turnStyle;
#ifndef WINVER
	int thing, i;
	NextHextrisWidget w = (NextHextrisWidget) renew;
#endif

#ifdef WINVER
	setValuesNextGame(w);
#else
	w->hextris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->hextris.mono);
	for (thing = 0; thing < MAX_START_POLYHEXES; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			 w->hextris.thingGC[thing][i] = NULL;
	}
	for (i = 0; i < BG_SHADES; i++)
		w->hextris.inverseGC[i] = NULL;
#endif

	w->hextris.quit = False;
	w->hextris.focus = False;
	w->hextris.next.units = w->hextris.units;
	w->hextris.next.corners = w->hextris.corners;
	w->hextris.next.mixed = w->hextris.mixed;
	w->hextris.nextThing.polyNumber = IGNORE_RAND;
	checkThings(w);
#if 0
	sizeNextGame(w);
#else
	resizeNextGame(w);
#endif
#ifdef WINVER
	brush = CreateSolidBrush(w->hextris.inverseGC[2]);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	setAllColors(w);
#endif
	corners = (w->hextris.next.corners) ? 1 : 0;
	mixed = (w->hextris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->hextris.next.units -
		MIN_UNITS].mode[corners].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "NextHextris: corrupted polyhex file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->hextris.numberPolyhexes = tris[w->hextris.next.units -
		MIN_UNITS].number[mixed][turnStyle];
	if (w->hextris.numberPolyhexes <= 0 ||
			w->hextris.numberPolyhexes > MAX_START_POLYHEXES) {
		(void) fprintf(stderr, "NextHextris: corrupted polyhex file\n");
		(void) fprintf(stderr, "\tnumberPolyhexes = %d.\n",
			w->hextris.numberPolyhexes);
		exit(1);
	}
}

#ifndef WINVER
static
#endif
void
exposeNextGame(
#ifdef WINVER
NextHextrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	NextHextrisWidget w = (NextHextrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	eraseFrame(w);
#if 0
	resizeThings(w);
#endif
	drawFrame(w);
	if (w->hextris.showNext) {
		drawThing(w);
	}
}

#ifndef WINVER
static
#endif
void
hideNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setNextGame(w, ACTION_HIDE);
}

#ifdef WINVER
void
valueNextGame(NextHextrisWidget w,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean corners, Boolean mixed, Boolean newBonus)
{
	w->hextris.nextThing.randomNumber = randomNumber;
	w->hextris.nextThing.randomRotation = randomRotation;
	w->hextris.nextThing.randomReflection = randomReflection;
	w->hextris.next.units = w->hextris.units = units;
	w->hextris.next.corners = w->hextris.corners = corners;
	w->hextris.next.mixed = w->hextris.mixed = mixed;
	w->hextris.nextThing.bonus = newBonus;
	redoNextNext(w);
	if (w->hextris.showNext) {
		drawThing(w);

	}
}
#endif

#ifndef WINVER
static
#endif
void
enterNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->hextris.focus = True;
	drawFrame(w);
}

#ifndef WINVER
static
#endif
void
leaveNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->hextris.focus = False;
	drawFrame(w);
}

#ifndef WINVER
static
#endif
void
showNextNextGame(NextHextrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->hextris.paused) {
		w->hextris.showNext = !w->hextris.showNext;
		if (w->hextris.showNext) {
			drawThing(w);
		} else {
			clearThing(w);
			drawFrame(w);
		}
		FLUSH(w);
	}
}
